# -*- coding: utf-8 -*-
""" peltak package. """
from __future__ import absolute_import, unicode_literals
from os.path import abspath, dirname

__version__ = '0.21.1'
PKG_DIR = abspath(dirname(__file__))



# -*- coding: utf-8 -*-
""" Code linting commands. """
from __future__ import absolute_import

from typing import List

# local imports
from . import root_cli, click


@root_cli.group('lint', invoke_without_command=True)
@click.option(
    '-e', '--exclude',
    multiple=True,
    metavar='PATTERN',
    help=("Specify patterns to exclude from linting. For multiple patterns, "
          "use the --exclude option multiple times")
)
@click.option(
    '--skip-untracked',
    is_flag=True,
    help="Also include files not tracked by git."
)
@click.option(
    '--commit', 'commit_only',
    is_flag=True,
    help=("Only lint files staged for commit. Useful if you want to clean up "
          "a large code base one commit at a time.")
)
@click.option(
    '-p', '--pretend',
    is_flag=True,
    help=("Just print files that would be linted without running anything")
)
@click.pass_context
def lint_cli(ctx, exclude, skip_untracked, commit_only, pretend):
    # type: (click.Context, List[str], bool, bool, bool) -> None
    """ Run pep8 and pylint on all project files.

    You can configure the linting paths using the lint.paths config variable.
    This should be a list of paths that will be linted. If a path to a directory
    is given, all files in that directory and it's subdirectories will be
    used.

    The pep8 and pylint config paths are by default stored in ops/tools/pep8.ini
    and ops/tools/pylint.ini. You can customise those paths in your config with
    lint.pep8_cfg and lint.pylint_cfg variables.

    **Config Example**::

        \b
        lint:
          pylint_cfg: 'ops/tools/pylint.ini'
          pep8_cfg: 'ops/tools/pep8.ini'
          paths:
            - 'src/mypkg'

    **Examples**::

        \b
        $ peltak lint               # Run linter in default mode, skip untracked
        $ peltak lint --commit      # Lint only files staged for commit
        $ peltak lint --all         # Lint all files, including untracked.
        $ peltak lint --pretend     # Print the list of files to lint
        $ peltak lint -e "*.tox*"   # Don't lint files inside .tox directory

    """
    if ctx.invoked_subcommand:
        return

    from peltak.commands import lint
    lint.lint(exclude, skip_untracked, commit_only, pretend)


# Used in docstrings only until we drop python2 support
del List



# -*- coding: utf-8 -*-
""" Application entry point. """
# pylint: disable=unused-import
from __future__ import absolute_import, unicode_literals

# local imports
from peltak.core import conf
from peltak.cli import root_cli
__all__ = [
    'root_cli'
]


conf.load()


from peltak.cli.root import clean   # noqa



# -*- coding: utf-8 -*-
"""
Per-project configuration support
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
import os
import os.path
import sys
from contextlib import contextmanager
from types import ModuleType
from typing import Any, Dict, Optional, Union

# 3rd party imports
import yaml

# local imports
from peltak import PKG_DIR
from . import util


PROJ_CONF_FILE = 'pelconf.py'
g_config = {}


def init(config):
    # type: (Dict[str, Any]) -> None
    """ Initialize configuration with the given values.

    This should be called from within the project fabfile, before any
    other commands are imported.

    This will not update the existing configuration but replace it entirely.

    Args:
        config (dict[str, Any]):
            The dictionary containing the project configuration.
    """
    global g_config

    g_config = config


def load():
    # type: () -> None
    """ Load configuration from file.

    This will search the directory structure upwards to find the project root
    (directory containing ``pelconf.py`` file). Once found it will import the
    config file which should initialize all the configuration (using
    `peltak.core.conf.init()` function).

    You can also have both yaml (configuration) and python (custom commands)
    living together. Just remember that calling `conf.init()` will overwrite
    the config defined in YAML.
    """
    with within_proj_dir():
        if os.path.exists('pelconf.yaml'):
            load_yaml_config('pelconf.yaml')

        if os.path.exists('pelconf.py'):
            load_py_config('pelconf.py')


def load_yaml_config(conf_file):
    # type: (str) -> None
    """ Load a YAML configuration.

    This will not update the configuration but replace it entirely.

    Args:
        conf_file (str):
            Path to the YAML config. This function will not check the file name
            or extension and will just crash if the given file does not exist or
            is not a valid YAML file.
    """
    global g_config

    with open(conf_file) as fp:
        # Initialize config
        g_config = yaml.load(fp)

        # Add src_dir to sys.paths if it's set. This is only done with YAML
        # configs, py configs have to do this manually.
        src_dir = get_path('src_dir', None)
        if src_dir is not None:
            sys.path.insert(0, src_dir)

        for cmd in get('commands', []):
            _import(cmd)


def load_py_config(conf_file):
    # type: (str) -> None
    """ Import configuration from a python file.

    This will just import the file into python. Sky is the limit. The file
    has to deal with the configuration all by itself (i.e. call conf.init()).
    You will also need to add your src directory to sys.paths if it's not the
    current working directory. This is done automatically if you use yaml
    config as well.

    Args:
        conf_file (str):
            Path to the py module config. This function will not check the file
            name or extension and will just crash if the given file does not
            exist or is not a valid python file.
    """
    if sys.version_info >= (3, 5):
        from importlib import util

        spec = util.spec_from_file_location('pelconf', conf_file)
        mod = util.module_from_spec(spec)
        spec.loader.exec_module(mod)

    elif sys.version_info >= (3, 3):
        from importlib import machinery
        loader = machinery.SourceFileLoader('pelconf', conf_file)
        _ = loader.load_module()

    elif sys.version_info <= (3, 0):
        import imp

        imp.load_source('pelconf', conf_file)


def load_template(filename):
    # type: (str) -> str
    """ Load template from file.

    The templates are part of the package and must be included as
    ``package_data`` in project ``setup.py``.

    Args:
        filename (str):
            The template path. Relative to `peltak` package directory.

    Returns:
        str: The content of the chosen template.
    """
    template_file = os.path.join(PKG_DIR, 'templates', filename)
    with open(template_file) as fp:
        return fp.read()


def _import(module):
    # type: (str) -> ModuleType
    return __import__(module)   # nocov


def getenv(name, default=None):
    # type: (str, Any) -> Union[str, Any]
    """ Get the value of an ENV variable. """
    return os.environ.get(name, default)   # nocov


def proj_path(path=None):
    # type: (str) -> str
    """ Return absolute path to the repo dir (root project directory).

    Args:
        path (str):
            The path relative to the project root (pelconf.yaml).

    Returns:
        str: The given path converted to an absolute path.
    """
    path = path or '.'

    if not os.path.isabs(path):
        proj_path = _find_proj_root()

        if proj_path is not None:
            path = os.path.normpath(os.path.join(proj_path, path))

    return path


@contextmanager
def within_proj_dir(path='.'):
    # type: (Optional[str]) -> str
    """ Return an absolute path to the given project relative path.

    :param path:
        Project relative path that will be converted to the system wide absolute
        path.
    :return:
        Absolute path.
    """
    curr_dir = os.getcwd()

    os.chdir(proj_path(path))

    yield

    os.chdir(curr_dir)


def get(name, *default):
    # type: (str, Any) -> Any
    """ Get config value with the given name and optional default.

    Args:
        name (str):
            The name of the config value.
        *default (Any):
            If given and the key doesn't not exist, this will be returned
            instead. If it's not given and the config value does not exist,
            AttributeError will be raised

    Returns:
        The requested config value. This is one of the global values defined
        in this file. If the value does not exist it will return `default` if
        give or raise `AttributeError`.

    Raises:
        AttributeError: If the value does not exist and `default` was not given.
    """
    global g_config

    curr = g_config
    for part in name.split('.'):
        if part in curr:
            curr = curr[part]
        elif default:
            return default[0]
        else:
            raise AttributeError("Config value '{}' does not exist".format(
                name
            ))

    return curr


def get_path(name, *default):
    # type: (str, Any) -> Any
    """ Get config value as path relative to the project directory.

    This allows easily defining the project configuration within the fabfile
    as always relative to that fabfile.

    Args:
        name (str):
            The name of the config value containing the path.
        *default (Any):
            If given and the key doesn't not exist, this will be returned
            instead. If it's not given and the config value does not exist,
            AttributeError will be raised

    Returns:
        The requested config value. This is one of the global values defined
        in this file. If the value does not exist it will return `default` if
        give or raise `AttributeError`.

    Raises:
        AttributeError: If the value does not exist and `default` was not given.
    """
    global g_config

    value = get(name, *default)

    if value is None:
        return None

    return proj_path(value)


@util.cached_result()
def _find_proj_root():
    # type: () -> Optional[str]
    """ Find the project path by going up the file tree.

    This will look in the current directory and upwards for the pelconf file
    (.yaml or .py)
    """
    proj_files = frozenset(('pelconf.py', 'pelconf.yaml'))
    curr = os.getcwd()

    while curr.startswith('/') and len(curr) > 1:
        if proj_files & frozenset(os.listdir(curr)):
            return curr
        else:
            curr = os.path.dirname(curr)

    return None


# Used in docstrings only until we drop python2 support
del Any, Dict, Optional, Union, ModuleType



# -*- coding: utf-8 -*-
""" Docker registry client. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
from typing import Any, Iterator, List, Tuple


class RegistryClient(object):
    """ Helper class for talking to docker registry.

    Attributes:
        user (str):
            Docker registry user name.
        pw (str):
            Docker registry password.
        auth (tuple[str, str]):
            Tuple of ``(self.user, self.pw)`` for easier access.
        registry_url (str):
            The remote registry URL.
    """
    def __init__(self, registry, username, password):
        self.user = username                        # type: str
        self.pw = password                          # type: str
        self.auth = (username, password)            # type: Tuple[str, str]
        self.registry_url = 'https://' + registry   # type: str

    def get(self, *args, **kw):
        # type: (*Any, *Any) -> requests.Response
        """ Proxy over requests.get

        Do not include requests globally so it's not required to build and
        push images. Otherwise, the project needs to install requests even
        when not using registry in any way.
        """
        import requests
        return requests.get(*args, **kw)

    def list_images(self):
        # type: () -> List[str]
        """ List images stored in the registry.

        Returns:
            list[str]: List of image names.
        """
        r = self.get(self.registry_url + '/v2/_catalog', auth=self.auth)
        return r.json()['repositories']

    def list_tags(self, image_name):
        # type: (str) -> Iterator[str]
        """ List all tags for the given image stored in the registry.

        Args:
            image_name (str):
                The name of the image to query. The image must be present on the
                registry for this call to return any values.
        Returns:
            list[str]: List of tags for that image.
        """
        tags_url = self.registry_url + '/v2/{}/tags/list'

        r = self.get(tags_url.format(image_name), auth=self.auth)
        data = r.json()

        if 'tags' in data:
            return reversed(sorted(data['tags']))

        return []


# Used in docstrings only until we drop python2 support
del Any, Iterator, List, Tuple



# -*- coding: utf-8 -*-
""" git flow feature commands. """
from peltak.cli import root_cli, click


@root_cli.group('feature', invoke_without_command=True)
def feature_cli():
    # type: () -> None
    """ Start a new git-flow feature.  """


@feature_cli.command('start')
@click.argument('name', required=False)
def start(name):
    # type: (str) -> None
    """ Start a new git-flow feature.  """
    from peltak.extra.gitflow import logic

    if name is None:
        name = click.prompt('Feature name')

    logic.feature.start(name)


@feature_cli.command('rename')
@click.argument('name', required=False)
def rename(name):
    # type: (str) -> None
    """ Give the currently developed feature a new name. """
    from peltak.extra.gitflow import logic

    if name is None:
        name = click.prompt('Feature name')

    logic.feature.rename(name)


@feature_cli.command('update')
def update():
    # type: () -> None
    """ Update the feature with updates committed to develop. """
    from peltak.extra.gitflow import logic
    logic.feature.update()


@feature_cli.command('finish')
def finish():
    # type: () -> None
    """ Merge current feature into develop. """
    from peltak.extra.gitflow import logic
    logic.feature.finish()


@feature_cli.command('merged')
def merged():
    # type: () -> None
    """ Cleanup a remotely merged branch. """
    from peltak.extra.gitflow import logic
    logic.feature.merged()



# -*- coding: utf-8 -*-
""" Commands package.

All commands (and only commands) should be defined inside this package. This
should be as thin layer as possible. Ideally just processing CLI params and
displaying results.
"""
from __future__ import absolute_import

# 3rd party imports
import click

# local imports
import peltak


@click.group()
@click.version_option(version=peltak.__version__, message='%(version)s')
def root_cli():
    # type: () -> None
    """

    To get help for a specific command:

       \033[1m peltak <command> --help\033[0m

    Examples:

       \033[1m peltak lint --help\033[0m

       \033[1m peltak version bump --help\033[0m

       \033[1m peltak release upload --help\033[0m

    """
    pass



# -*- coding: utf-8 -*-
""" Docker commands implementation. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
import sys
from typing import Any, Dict

# 3rd party imports
import click

# local imports
from peltak.core import shell
from peltak.core import log
from peltak.core import conf
from peltak.core import versioning
from . import client


def build_images():
    # type: () -> None
    """ Build all docker images for the project. """
    registry = conf.get('docker.registry')
    docker_images = conf.get('docker.images', [])

    for image in docker_images:
        build_image(registry, image)


def push_images():
    # type: () -> None
    """ Push all project docker images to a remote registry. """
    registry = conf.get('docker.registry')
    docker_images = conf.get('docker.images', [])

    if registry is None:
        log.err("You must define docker.registry conf variable to push images")
        sys.exit(-1)

    for image in docker_images:
        push_image(registry, image)


def docker_list(registry_pass):
    # type: (str) -> None
    """ List docker images stored in the remote registry.

    Args:
        registry_pass (str):
            Remote docker registry password.
    """
    registry = conf.get('docker.registry', None)

    if registry is None:
        log.err("You must define docker.registry conf variable to list images")
        sys.exit(-1)

    registry_user = conf.get('docker.registry_user', None)

    if registry_user is None:
        registry_user = click.prompt("Username")

    rc = client.RegistryClient(registry, registry_user, registry_pass)
    images = {x: rc.list_tags(x) for x in rc.list_images()}

    shell.cprint("<32>Images in <34>{} <32>registry:", registry)
    for image, tags in images.items():
        shell.cprint('  <92>{}', image)
        for tag in tags:
            shell.cprint('      <90>{}:<35>{}', image, tag)


def build_image(registry, image):
    # type: (str, Dict[str, Any]) -> None
    """ Build docker image.

    Args:
        registry (str):
            The name of the registry this image belongs to. If not given, the
            resulting image will have a name without the registry.
        image (dict[str, Any]):
            The dict containing the information about the built image. This is
            the same dictionary as defined in DOCKER_IMAGES variable.
    """
    if ':' in image['name']:
        _, tag = image['name'].split(':', 1)
    else:
        _, tag = image['name'], None

    values = {
        'registry': '' if registry is None else registry + '/',
        'image': image['name'],
        'tag': tag,
    }

    if tag is None:
        args = [
            '-t {registry}{image}'.format(**values),
            '-t {registry}{image}:{version}'.format(
                version=versioning.current(),
                **values
            ),
        ]
    else:
        args = ['-t {registry}{image}'.format(**values)]

    if 'file' in image:
        args.append('-f {}'.format(conf.proj_path(image['file'])))

    with conf.within_proj_dir(image.get('path', '.')):
        log.info("Building <33>{registry}<35>/{image}", **values)
        shell.run('docker build {args} .'.format(args=' '.join(args)))


def push_image(registry, image):
    # type: (str, Dict[str, Any]) -> None
    """ Push the given image to selected repository.

    Args:
        registry (str):
            The name of the registry we're pushing to. This is the address of
            the repository without the protocol specification (no http(s)://)
        image (dict[str, Any]):
            The dict containing the information about the image. This is the
            same dictionary as defined in DOCKER_IMAGES variable.
    """
    values = {
        'registry': registry,
        'image': image['name'],
    }

    log.info("Pushing <33>{registry}<35>/{image}".format(**values))
    shell.run('docker push {registry}/{image}'.format(**values))


# Used in docstrings only until we drop python2 support
del Any, Dict



# -*- coding: utf-8 -*-
""" AppEngine commands implementation. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
import sys
from fnmatch import fnmatch
from os.path import exists, join, normpath

# 3rd party imports
import attr

# local imports
from peltak.core import conf
from peltak.core import fs
from peltak.core import git
from peltak.core import log
from peltak.core import shell
from peltak.core import versioning
from peltak.core import util


@util.mark_experimental
def deploy(app_id, version, pretend, promote, quiet):
    # type: (str, str, bool, bool, bool) -> None
    """ Deploy the app to AppEngine.

    Args:
        app_id (str):
            AppEngine App ID. Overrides config value app_id if given.
        version (str):
            AppEngine project version. Overrides config values if given.
        pretend (bool):
            If set to **True**, do not actually deploy anything but show the
            deploy command that would be used.
        promote (bool):
            If set to **True** promote the current remote app version to the one
            that's being deployed.
        quiet (bool):
            If set to **True** this will pass the ``--quiet`` flag to gcloud
            command.
    """
    gae_app = GaeApp.for_branch(git.current_branch().name)

    if gae_app is None and None in (app_id,  version):
        msg = (
            "Can't find an AppEngine app setup for branch <35>{}<32> and"
            "--project and --version were not given."
        )
        log.err(msg, git.current_branch().name)
        sys.exit(1)

    if version is not None:
        gae_app.version = version

    if app_id is not None:
        gae_app.app_id = app_id

    gae_app.deploy(promote, pretend, quiet)


@util.mark_experimental
def devserver(port, admin_port, clear):
    # type: (int, int, bool) -> None
    """ Run devserver.

    Args:
        port (int):
            Port on which the app will be served.
        admin_port (int):
            Port on which the admin interface is served.
        clear (bool):
            If set to **True**, clear the datastore on startup.
    """
    admin_port = admin_port or (port + 1)

    args = [
        '--port={}'.format(port),
        '--admin_port={}'.format(admin_port)
    ]

    if clear:
        args += ['--clear_datastore=yes']

    with conf.within_proj_dir():
        shell.run('dev_appserver.py . {args}'.format(args=' '.join(args)))


@util.mark_experimental
def setup_ci():
    # type: () -> None
    """ Setup AppEngine SDK on CircleCI """
    gcloud_path = shell.run('which gcloud', capture=True).stdout.strip()
    sdk_path = normpath(join(gcloud_path, '../../platform/google_appengine'))
    gcloud_cmd = gcloud_path + ' --quiet'

    if not exists(sdk_path):
        log.info("Installing AppEngine SDK")
        shell.run('sudo {} components install app-engine-python'.format(
            gcloud_cmd
        ))
    else:
        # Only initialise once. To reinitialise, just build without cache.
        log.info("AppEngine SDK already initialised")

    log.info("Using service account authentication")
    shell.run('{} auth activate-service-account --key-file {}'.format(
        gcloud_cmd,
        conf.proj_path('ops/client_secret.json')
    ))

    # shell.run('{} config set project {}'.format(gcloud_cmd, project))


def _find_appengine_sdk():
    # type: () -> str
    gcloud_path = shell.run('which gcloud', capture=True).stdout.strip()
    return normpath(join(gcloud_path, '../../platform/google_appengine'))


@attr.s
class GaeApp(object):
    """ Represents an AppEngine app.

    Attributes:
        app_id (str):
            AppEngine project ID.
        version (str):
            AppEngine project version.
        deployables (list[str]):
            List of YAML files to deploy. Defaults to all yaml files in the
            project root directory.
    """
    app_id = attr.ib(type=str)
    version = attr.ib(type=str, default=None)
    deployables = attr.ib(type=list, default=['.'])

    @classmethod
    def for_branch(cls, branch_name):
        # type: (str) -> GaeApp
        """ Return app configuration for the given branch.

        This will look for the configuration in the ``appengine.projects``
        config variable.

        Args:
            branch_name (str):
                The name of the branch we want the configuration for.

        Returns:
            Optional[GaeApp]: The `GaeApp` instance with the configuration or
            **None** if none found.
        """
        for proj in conf.get('appengine.projects', []):
            if fnmatch(branch_name, proj['branch']):
                app_config = dict(proj)
                app_config.pop('branch')
                return cls(**app_config)

        return None

    @property
    def app_version(self):
        # type: () -> str
        """ AppEngine compatible app version.

        This assumes git-flow branching strategy and has a few predefined ways
        to automatically generate the version string.
        """
        branch = git.current_branch()

        if self.version is not None:
            return self.version

        elif branch.type == 'develop':
            return '{ver}-c{commit_nr}-{commit_id}'.format(
                ver=versioning.current().replace('.', '-'),
                commit_nr=git.latest_commit().number,
                commit_id=git.latest_commit().id
            )

        elif branch.type in ('feature', 'hotfix', 'task'):
            # We don't need version when we have the title.
            return branch.title.replace('_', '-')

        return versioning.current().replace('.', '-')

    def deploy(self, promote=False, pretend=False, quiet=False):
        # type: (bool, bool, bool) -> None
        """ Deploy the code to AppEngine.

        Args:
            promote (bool):
                Migrate the traffic to the deployed version.
            pretend (bool):
                Instead of deploying, print the deploy command.
            quiet (bool):
                Pass ``--quiet`` flag to gcloud command
        """
        args = [
            '--promote' if promote else '--no-promote',
            '--version {}'.format(self.app_version),
            '--project {}'.format(self.app_id),
        ]

        if quiet:
            args += ['--quiet']

        cmd = 'gcloud app deploy {args} {deployables}'.format(
            deployables=fs.surround_paths_with_quotes(self.deployables),
            args=' '.join(args)
        )

        if pretend:
            log.info("Would deploy version <35>{ver}<32> to <35>{app}".format(
                ver=self.app_version,
                app=self.app_id
            ))
            shell.cprint('<90>{}', cmd)

        if not pretend:
            log.info("Deploying version <35>{ver}<32> to <35>{app}".format(
                ver=self.app_version,
                app=self.app_id,
            ))
            shell.run(cmd)



# -*- coding: utf-8 -*-
""" Miscellaneous commands implementation. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
import os
from os.path import exists, isdir, join
from shutil import rmtree
from typing import List

# 3rd party imports
import click
import cliform

# local imports
from peltak.core import conf
from peltak.core import fs
from peltak.core import log
from peltak.core import shell
from peltak.core import util


def clean(pretend, exclude):
    # type: (bool, List[str]) -> None
    """ Remove all unnecessary files.

    Args:
        pretend (bool):
            If set to **True**, do not delete any files, just show what would be
            deleted.
        exclude (list[str]):
            A list of path patterns to exclude from deletion.
    """
    exclude = list(exclude) + conf.get('clean.exclude', [])
    clean_patterns = conf.get('clean.patterns', [
        '*__pycache__*',
        '*.py[cod]',
        '*.swp',
    ])

    num_files = 0
    with util.timed_block() as t:
        files = fs.filtered_walk(conf.proj_path(), clean_patterns, exclude)
        for path in files:
            try:
                num_files += 1

                if not isdir(path):
                    log.info('  <91>[file] <90>{}', path)
                    not pretend and os.remove(path)
                else:
                    log.info('  <91>[dir]  <90>{}', path)
                    not pretend and rmtree(path)

            except OSError:
                log.info("<33>Failed to remove <90>{}", path)

    if pretend:
        msg = "Would delete <33>{}<32> files. Took <33>{}<32>s"
    else:
        msg = "Deleted <33>{}<32> files in <33>{}<32>s"

    log.info(msg.format(num_files, t.elapsed_s))


class InitForm(cliform.Form):
    """ Everything needed to generate initial pelconf.yaml. """
    src_dir = cliform.Field(
        'Source directory',
        default='src',
        type=str,
        help=('The root directory for all your sources. This is what you '
              'would treat as PYTHONPATH')
    )
    build_dir = cliform.Field(
        'Build directory',
        default='.build',
        type=str,
        help=('The directory where all build files should go. It will be '
              'shared as much as possible across the tools to reduce '
              'clutter')
    )
    src_path = cliform.Field(
        'Path to your main package',
        type=str,
        help=('This will be used for a lot of default values in the config '
              '(like for linting, reference docs etc.)')
    )
    version_file = cliform.Field(
        'Version file',
        type=str,
        default=lambda f: join(f['src_path'], '__init__.py'),
        help=('This will be used for a lot of default values in the config '
              '(like for linting, reference docs etc.)')
    )


def init(quick):
    # type: () -> None
    """ Create an empty pelconf.yaml from template """
    config_file = 'pelconf.yaml'
    prompt = "-- <35>{} <32>already exists. Wipe it?<0>".format(config_file)

    if exists(config_file) and not click.confirm(shell.fmt(prompt)):
        log.info("Canceled")
        return

    form = InitForm().run(quick=quick)

    log.info('Writing <35>{}'.format(config_file))
    with open(config_file, 'w') as fp:
        pelconf_template = conf.load_template('pelconf.yaml')
        fp.write(pelconf_template.format(**form.values))


# Used in docstrings only until we drop python2 support
del List



# -*- coding: utf-8 -*-
""" Helper commands for dealing with pypi. """
from __future__ import absolute_import, unicode_literals
from peltak.cli import root_cli, click


@root_cli.group('pypi')
def pypi_cli():
    # type: () -> None
    """ pypi related commands """
    pass


@pypi_cli.command('upload')
@click.argument('target')
def upload(target):
    # type: (str) -> None
    """ Upload to a given pypi target.

    Examples::

        \b
        $ peltak pypi upload pypi    # Upload the current code to pypi
        $ peltak pypi upload private # Upload to pypi server 'private'

    """
    from peltak.extra.pypi import logic
    logic.upload(target)


@pypi_cli.command('configure')
@click.option(
    '-u', '--username',
    type=str,
    help="PyPi username. Defaults to PYPI_USER env variable."
)
@click.option(
    '-p', '--password',
    type=str,
    help="PyPi password. Defaults to PYPI_PASS env variable."
)
def configure(username, password):
    # type: (str, str) -> None
    """
    Generate .pypirc config with the given credentials.

    Example:

        $ peltak pypi configure my_pypi_user my_pypi_pass

    """
    from peltak.extra.pypi import logic
    logic.gen_pypirc(username, password)



# -*- coding: utf-8 -*-
""" git helpers. """
from __future__ import absolute_import
from . import root_cli


@root_cli.group('git')
def git_cli():
    # type: () -> None
    """ Git related commands """
    pass


@git_cli.command('add-hooks')
def add_hooks():
    # type: () -> None
    """ Setup project git hooks.

    This will run all the checks before pushing to avoid waiting for the CI
    fail.

    Example::

        $ peltak git add-hooks
    """
    from peltak.commands import git

    git.add_hooks()


@git_cli.command('push')
def push():
    # type: () -> None
    """ Push the current branch and set to track remote.

    Example::

        $ peltak git push

    """
    from peltak.commands import git

    git.push()



# -*- coding: utf-8 -*-
""" Simple command line form/wizard implementation. """
from __future__ import absolute_import, unicode_literals
from .core import Field, Form
__all__ = [
    'Field',
    'Form'
]



# -*- coding: utf-8 -*-
""" git flow hotfix commands implementation. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
import sys

# local imports
from peltak.core import conf
from peltak.core import git
from peltak.core import log
from . import common


def start(name):
    # type: (str) -> None
    """ Start working on a new hotfix.

    This will create a new branch off master called hotfix/<name>.

    Args:
        name (str):
            The name of the new feature.
    """
    hotfix_branch = 'hotfix/' + common.to_branch_name(name)
    master = conf.get('git.master_branch', 'master')

    common.assert_on_branch(master)
    common.git_checkout(hotfix_branch, create=True)


def rename(name):
    # type: (str) -> None
    """ Give the currently developed feature a new name.

    Args:
        name (str):
            The new name of the current feature. The current branch will be
            renamed to 'feature/<new_name>'.
    """
    common.assert_branch_type('hotfix')
    common.git_branch_rename('hotfix/' + name.strip().replace(' ', '_'))


def update():
    # type: () -> None
    """ Update the hotfix with updates committed to master.

    This will merge current develop into the current branch.
    """
    branch = git.current_branch(refresh=True)
    master = conf.get('git.master_branch', 'master')

    common.assert_branch_type('feature')
    common.git_checkout(master)
    common.git_pull(master)
    common.git_checkout(branch.name)
    common.git_merge(branch.name, master)


def finish():
    # type: () -> None
    """ Merge current feature into develop. """
    if git.staged() or git.unstaged():
        log.err(
            "You have uncommitted changes in your repo!\n"
            "You need to stash them before you merge the hotfix branch"
        )
        sys.exit(1)

    develop = conf.get('git.devel_branch', 'develop')
    master = conf.get('git.master_branch', 'master')
    branch = git.current_branch(refresh=True)

    common.assert_branch_type('hotfix')

    # Merge hotfix into master
    common.git_checkout(master)
    common.git_pull(master)
    common.git_merge(master, branch.name)

    # Merge hotfix into develop
    common.git_checkout(develop)
    common.git_pull(develop)
    common.git_merge(develop, branch.name)

    # Cleanup
    common.git_branch_delete(branch.name)
    common.git_prune()

    common.git_checkout(master)


def merged():
    # type: () -> None
    """ Cleanup a remotely merged branch. """
    develop = conf.get('git.devel_branch', 'develop')
    master = conf.get('git.master_branch', 'master')
    branch = git.current_branch(refresh=True)

    common.assert_branch_type('hotfix')

    # Pull master with the merged hotfix
    common.git_checkout(master)
    common.git_pull(master)

    # Merge to develop
    common.git_checkout(develop)
    common.git_pull(develop)
    common.git_merge(develop, branch.name)

    # Cleanup
    common.git_branch_delete(branch.name)
    common.git_prune()

    common.git_checkout(master)



# -*- coding: utf-8 -*-
""" Commands for building and pushing the app image.

This is mainly for web apps that are deployed with docker.
"""
from __future__ import absolute_import
from .cli import docker_cli



# -*- coding: utf-8 -*-
""" Various helpers that do not depend on anything else in the project. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
import time
import warnings
from functools import wraps
from types import FunctionType


class timed_block(object):  # noqa
    """ Context manager to measure execution time for a give block of code.

    Example:
        >>> import time
        >>> from peltak.core import util
        >>>
        >>> with util.timed_block() as t:
        ...     time.sleep(1)
        >>>
        >>> print("Code executed in {}s".format(int(t.elapsed_s)))
        Code executed in 1s

    Attributes:
        t0 (float):
            The time at the start of execution.
        elapsed (float);
            Raw elapsed time in seconds.
        elapsed_s (float):
            Elapsed time in seconds, rounded to 3 decimal places for easy
            display.
        elapsed_ms (float):
            Elapsed time in milliseconds, rounded to 3 decimal places for easy
            display.
    """
    def __enter__(self):
        self.t0 = time.time()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.elapsed = time.time() - self.t0
        self.elapsed_s = round(self.elapsed, 3)
        self.elapsed_ms = round(self.elapsed * 1000, 3)


def mark_experimental(fn):
    # type: (FunctionType) -> FunctionType
    """ Mark function as experimental.

    Args:
        fn (FunctionType):
            The command function to decorate.
    """
    @wraps(fn)
    def wrapper(*args, **kw):   # pylint: disable=missing-docstring
        warnings.warn("This command is has experimental status. The interface "
                      "is not yet stable and might change without notice "
                      "within with a patch version update. "
                      "Use at your own risk")
        return fn(*args, **kw)

    return wrapper


def mark_deprecated(replaced_by):
    # type: (str) -> FunctionType
    """ Mark command as deprecated.

    Args:
        replaced_by (str):
            The command that deprecated this command and should be used instead.
    """
    def decorator(fn):   # pylint: disable=missing-docstring
        @wraps(fn)
        def wrapper(*args, **kw):   # pylint: disable=missing-docstring
            warnings.warn("This command is has been deprecated. Please use "
                          "{new} instead.".format(new=replaced_by))

            return fn(*args, **kw)

        return wrapper

    return decorator


class cached_result(object):
    """ Decorator that caches the function result.

    This is especially useful for functions whose output won't change during
    a single *peltak* execution run and are expensive (i.e. call external
    shell commands).

    Example:
        >>> from peltak.core import util
        >>>
        >>> @util.cached_result()
        ... def foo():
        ...     call_count = getattr(foo, 'call_count', 0)
        ...     call_count += 1
        ...     setattr(foo, 'call_count', call_count)
        >>> foo()
        >>> print(foo.call_count)
        1
        >>> foo()
        >>> print(foo.call_count)
        1
        >>> util.cached_result.clear(foo)
        >>> foo()
        >>> print(foo.call_count)
        2
    """
    CACHE_VAR = '__cached_result'

    def __init__(self):
        pass

    def __call__(self, fn):
        # type: (FunctionType) -> FunctionType
        """ Apply the decorator to the given function.

        Args:
            fn (FunctionType):
                The function to decorate.
        :return Function:
            The function wrapped in caching logic.
        """
        @wraps(fn)
        def wrapper(refresh=False):   # pylint: disable=missing-docstring
            if refresh or not hasattr(wrapper, self.CACHE_VAR):
                result = fn()
                setattr(wrapper, self.CACHE_VAR, result)

            return getattr(wrapper, self.CACHE_VAR)

        return wrapper

    @classmethod
    def clear(cls, fn):
        # type: (FunctionType) -> None
        """ Clear result cache on the given function.

        If the function has no cached result, this call will do nothing.

        Args:
            fn (FunctionType):
                The function whose cache should be cleared.
        """
        if hasattr(fn, cls.CACHE_VAR):
            delattr(fn, cls.CACHE_VAR)


# Used in docstrings only until we drop python2 support
del FunctionType



# -*- coding: utf-8 -*-
""" Documentation commands implementation. """
from __future__ import absolute_import, unicode_literals, print_function

# stdlib imports
import os.path
import shutil
import sys

# local imports
from peltak.core import conf
from peltak.core import log
from peltak.core import shell


def docs(recreate, gen_index, run_doctests, verbose):
    # type: (bool, bool, bool, int) -> None
    """ Build the documentation for the project.

    Args:
        recreate (bool):
            If set to **True**, the build and output directories will be cleared
            prior to generating the docs.
        gen_index (bool):
            If set to **True**, it will generate top-level index file for the
            reference documentation.
        run_doctests (bool):
            Set to **True** if you want to run doctests after the documentation
            is generated.
        verbose (int):
            The verbosity level.
    """
    build_dir = conf.get_path('build_dir', '.build')
    docs_dir = conf.get_path('docs.path', 'docs')
    refdoc_paths = conf.get('docs.reference', [])

    docs_html_dir = os.path.join(docs_dir, 'html')
    docs_tests_dir = os.path.join(docs_dir, 'doctest')
    docs_build_dir = os.path.join(build_dir, 'docs')

    if recreate:
        for path in (docs_html_dir, docs_build_dir):
            if os.path.exists(path):
                log.info("<91>Deleting <94>{}".format(path))
                shutil.rmtree(path)

    if refdoc_paths:
        gen_ref_docs(verbose, gen_index)
    else:
        log.err('Not generating any reference documentation - '
                'No docs.reference specified in config')

    with conf.within_proj_dir(docs_dir):
        log.info('Building docs')
        shell.run('sphinx-build -b html -d {build} {docs} {out}'.format(
            build=docs_build_dir,
            docs=docs_dir,
            out=docs_html_dir,
        ))

        if run_doctests:
            log.info('Running doctests')
            shell.run('sphinx-build -b doctest -d {build} {docs} {out}'.format(
                build=docs_build_dir,
                docs=docs_dir,
                out=docs_tests_dir,
            ))

        log.info('You can view the docs by browsing to <34>file://{}'.format(
            os.path.join(docs_html_dir, 'index.html')
        ))


def gen_ref_docs(verbose, gen_index=False):
    # type: (int, bool) -> None
    """ Generate reference documentation for the project.

    This will use **sphinx-refdoc** to generate the source .rst files for the
    reference documentation.

    Args:

        verbose (int):
            Verbosity level. This will be passed directly to sphinx-refdoc.
        gen_index (bool):
            Set it to **True** if you want to generate the index file with the
            list of top-level packages. This is set to default as in most cases
            you only have one package per project so you can link directly to
            that package reference (and if index were generated sphinx would
            complain about file not included in toctree).
    """
    try:
        from refdoc import generate_docs
    except ImportError as ex:
        msg = ("You need to install sphinx-refdoc if you want to generate "
               "code reference docs.")

        print(msg, file=sys.stderr)
        log.err("Exception: {}".format(ex))
        sys.exit(-1)

    docs_dir = conf.get_path('docs.path', 'docs')
    docs_ref_dir = os.path.join(docs_dir, 'ref')
    refdoc_paths = conf.get('docs.reference', [])

    log.info('Removing previously generated reference documentation')
    if os.path.exists(docs_ref_dir):
        shutil.rmtree(docs_ref_dir)

    log.info('Generating reference documentation')
    args = {
        'out_dir': docs_ref_dir,
        'verbose': verbose,
    }

    if gen_index:
        args['gen_index'] = True

    pkg_paths = [conf.proj_path(p) for p in refdoc_paths]

    generate_docs(pkg_paths, **args)



# -*- coding: utf-8 -*-
""" git flow release commands. """
from peltak.cli import root_cli, click


@root_cli.group('release', invoke_without_command=True)
def release_cli():
    # type: () -> None
    """ Commands that implement the release git flow.

    **Example Config**::

        \b
        version_file: 'src/mypkg/__init__.py'
        git:
            devel_branch: 'develop'
            master_branch: 'master'
            protected_branches: ['master', 'develop']

    Examples:

        \b
        $ peltak release start patch    # Make a new patch release
        $ peltak release start minor    # Make a new minor release
        $ peltak release start major    # Make a new major release
        $ peltak release start          # same as start patch
        $ peltak release tag            # Tag current commit as release

    """
    pass


@release_cli.command('start')
@click.argument(
    'component',
    type=click.Choice(['major', 'minor', 'patch']),
    default='patch',
)
@click.option(
    '--exact',
    type=str,
    help="Set the newly released version to be exactly as specified."
)
def start(component, exact):
    # type: (str) -> None
    """ Create a new release.

    It will bump the current version number and create a release branch called
    `release/<version>` with one new commit (the version bump).

    **Example Config**::

        \b
        version_file: 'src/mypkg/__init__.py'

    **Examples**::

        \b
        $ peltak release start patch    # Make a new patch release
        $ peltak release start minor    # Make a new minor release
        $ peltak release start major    # Make a new major release
        $ peltak release start          # same as start patch

    """
    from peltak.extra.gitflow import logic
    logic.release.start(component, exact)


@release_cli.command('tag')
def tag_release():
    # type: () -> None
    """ Tag the current commit with as the current version release.

    This should be the same commit as the one that's uploaded as the release
    (to pypi for example).

    **Example Config**::

        \b
        version_file: 'src/mypkg/__init__.py'

    Examples::

        $ peltak release tag          # Tag the current commit as release

    """
    from peltak.extra.gitflow import logic
    logic.release.tag()


@release_cli.command('finish')
def finish():
    # type: () -> None
    """ Merge the current release to both develop and master.

    This will perform a FF merge with develop if possible and --no-ff merge
    with master and then tag the merge commit with the current version.
    """
    from peltak.extra.gitflow import logic
    logic.release.finish()


@release_cli.command('merged')
def merged():
    # type: () -> None
    """ Checkout the target branch, pull and delete the merged branch.

    This is to ease the repetitive cleanup of each merged branch.

    Example Config (those the are defaults)::

        \b
        git:
            devel_branch: 'develop'
            master_branch: 'master'
            protected_branches: ['master', 'develop']

    Example::

        $ peltak release merged     # Must be ran on the relase branch

    """
    from peltak.extra.gitflow import logic
    logic.release.merged()



# -*- coding: utf-8 -*-
""" Helpers for testing peltak related code.

This is kept inside the peltak package as it ca be used by 3rd party packages
extending the app.
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
import os
import shutil
import tempfile
from contextlib import contextmanager
from functools import wraps
from os.path import basename, join
from types import FunctionType
from typing import Any, Dict, List, Union

# 3rd party imports
from mock import patch, Mock
from peltak.core.shell import ExecResult


class patch_proj_root(object):
    """ Patch project root decorator. """
    def __init__(self, proj_root, nest_level=0):
        self.proj_root = proj_root
        self.nest = nest_level

    def __call__(self, fn):
        if self.proj_root is not None:
            cwd = join(self.proj_root, *(['fake_dir'] * self.nest))
            dirs = ['not_pelconf'] * self.nest + ['pelconf.py']
            dirs = [[d] for d in dirs]
        else:
            cwd = join('/', *(['fake_dir'] * self.nest))
            dirs = [[d] for d in ['not_pelconf'] * self.nest]

        @patch('os.getcwd', Mock(return_value=cwd))
        @patch('os.listdir', Mock(side_effect=dirs))
        @wraps(fn)
        def wrapper(*args, **kw):       # pylint: disable=missing-docstring
            return fn(*args, **kw)

        return wrapper


def patch_pelconf(config):
    # type: (Dict[str, Any]) -> FunctionType
    """ Patch the peltak configuration.

    This will patch all content retrieved through `peltak.core.conf.get()` and
    `conf.get_path()`.

    Args:
        config (dict[str, Any]):
            The dictionary with the peltak configuration.
    """
    return patch('peltak.core.conf.g_config', config)


def mock_response(data):
    # type: (Union[Dict[str, Any], List[Any]]) -> Mock
    """ Create a mock response that can be set as requests return value.

    Args:
        data (Union[dict[str, Any], list[Any]]):

    Returns:
        MagicMock:
    """
    resp_mock = Mock()
    resp_mock.json = Mock(return_value=data)
    return resp_mock


class TestDataProvider(object):
    """ Provides a temporary copy of the selected test project. """

    def __init__(self, data_dir):
        self.data_dir = data_dir
        self.proj_dir = None

    @contextmanager
    def __call__(self, path):
        temp_dir = tempfile.mkdtemp(prefix='peltak_tests_')

        cwd = os.getcwd()
        self.proj_dir = join(temp_dir, basename(path))
        proj_src = join(self.data_dir, path)

        shutil.copytree(proj_src,  self.proj_dir)
        os.chdir(self.proj_dir)

        try:
            yield self
        finally:
            os.chdir(cwd)
            shutil.rmtree(temp_dir)


def mock_result(stdout=None, retcode=None, stderr=None, cmd=None):
    # type: (str, int, str, str) -> ExecResult
    """ Helper for creating ExecResults for tests.

    Args:
        stdout (str):
        retcode (int):
        stderr (str):
        cmd (str):

    Returns:
        ExecResult:
    """
    return ExecResult(
        cmd or '',
        retcode or 0,
        stdout or '',
        stderr or '',
        retcode == 0,
        retcode != 0,
    )


def patch_run(stdout=None, retcode=None, stderr=None, cmd=None):
    # type: (str, int, str, str) -> FunctionType
    """ Patch shell.run and make it return a given result.

    Args:
        stdout (str):
        retcode (int):
        stderr (str):
        cmd (str):

    Returns:
        FunctionType:

    """
    p_run = Mock(return_value=ExecResult(
        cmd or '',
        retcode or 0,
        stdout or '',
        stderr or '',
        retcode == 0,
        retcode != 0,
    ))

    return patch('peltak.core.shell.run', p_run)



# -*- coding: utf-8 -*-
""" git flow hotfix commands. """
from peltak.cli import root_cli, click


@root_cli.group('hotfix', invoke_without_command=True)
def hotfix_cli():
    # type: () -> None
    """ Commands that ease the work with git flow hotfix branches.

    Examples:

        \b
        $ peltak hotfix start my_hotfix     # Start a new hotfix branch
        $ peltak hotfix finish              # Merge the hotfix into master
        $ peltak hotfix merged              # Cleanup after a remote merge
        $ peltak hotfix rename new_name     # Rename the current hotfix.
    """
    pass


@hotfix_cli.command('start')
@click.argument('name', required=False)
def start(name):
    # type: (str) -> None
    """ Start a new git flow hotfix branch.  """
    from peltak.extra.gitflow import logic

    if name is None:
        name = click.prompt('Hotfix name')

    logic.hotfix.start(name)


@hotfix_cli.command('rename')
@click.argument('name', required=False)
def rename(name):
    # type: (str) -> None
    """ Give the currently developed hotfix a new name. """
    from peltak.extra.gitflow import logic

    if name is None:
        name = click.prompt('Hotfix name')

    logic.hotfix.rename(name)


@hotfix_cli.command('update')
def update():
    # type: () -> None
    """ Update the hotfix with updates committed to master. """
    from peltak.extra.gitflow import logic
    logic.hotfix.update()


@hotfix_cli.command('finish')
def finish():
    # type: () -> None
    """ Merge current hotfix into master. """
    from peltak.extra.gitflow import logic
    logic.hotfix.finish()


@hotfix_cli.command('merged')
def merged():
    # type: () -> None
    """ Cleanup a remotely merged hotfix. """
    from peltak.extra.gitflow import logic
    logic.hotfix.merged()



# -*- coding: utf-8 -*-
""" Root level CLI commands. """
from __future__ import absolute_import

# stdlib imports
from typing import List

# local imports
from . import root_cli, click


@root_cli.command('clean')
@click.option(
    '-p', '--pretend',
    is_flag=True,
    help=("Just print files that would be deleted, without actually "
          "deleting them")
)
@click.option(
    '-e', '--exclude',
    multiple=True,
    metavar='PATTERN',
    help='Comma separated list of paths to exclude from deletion'
)
def clean(pretend, exclude):
    # type: (bool, List[str]) -> None
    """ Remove temporary files like python cache, swap files, etc.

    You can configure the list of patterns with clean_patterns config variable.
    By default it will remove all files/dirs matching

    Config example::

        \b
        clean:
          patterns:
            - '*__pycache__*',
            - '*.py[cod]',
            - '*.swp'
          exclude:
            - '.tox'
            - '.venv'

    Examples::

        \b
        $ peltak clean
        $ peltak clean -e "*.tox*"
        $ peltak clean --pretend

    """
    from peltak.commands import root
    root.clean(pretend, exclude)


@root_cli.command('init')
@click.option(
    '-q', '--quick',
    is_flag=True,
    help="Enable quick mode. Defaults will be used wherever possible."
)
def init(quick):
    # type: () -> None
    """ Create new peltak config file in the current directory.

    If ``pelconf.py`` already exists the user will be prompted to confirm
    before continuing.

    Example::

        $ peltak init

    """
    from peltak.commands import root
    root.init(quick)


# Used in docstrings only until we drop python2 support
del List
